O_TWO_CLASS = "Two classes"
O_THREE_CLASS = "Three classes"
O_WEIGHTED_VARIANCE = "Weighted variance"
O_ENTROPY = "Entropy"
O_FOREGROUND = "Foreground"
O_BACKGROUND = "Background"
FI_IMAGE_SIZE = "Image size"
FI_CUSTOM = "Custom"
RB_DEFAULT = "Default"
RB_CUSTOM = "Custom"
RB_MEAN = "Mean"
RB_MEDIAN = "Median"
RB_MODE = "Mode"
RB_SD = "Standard deviation"
RB_MAD = "Median absolute deviation"
TS_AUTOMATIC = "Automatic"
TS_GLOBAL = "Global"
TS_ADAPTIVE = "Adaptive"
TS_PER_OBJECT = "Per object"
TS_MANUAL = "Manual"
TS_BINARY_IMAGE = "Binary image"
TS_MEASUREMENT = "Measurement"
TS_ALL = [TS_GLOBAL, TS_ADAPTIVE, TS_MANUAL, TS_MEASUREMENT]
O_FROM_IMAGE = "From image"
TSM_NONE = "No smoothing"
TSM_AUTOMATIC = "Automatic"
TSM_MANUAL = "Manual"
TECH_NOTE_ICON = "gear.png"
