X_AUTOMATIC_EXTRACTION = "Extract from image file headers"
X_MANUAL_EXTRACTION = "Extract from file/folder names"
X_IMPORTED_EXTRACTION = "Import from file"
X_ALL_EXTRACTION_METHODS = [
    X_MANUAL_EXTRACTION,
    X_IMPORTED_EXTRACTION,
]
XM_FILE_NAME = "File name"
XM_FOLDER_NAME = "Folder name"
DTC_TEXT = "Text"
DTC_CHOOSE = "Choose for each"
DTC_ALL = [DTC_TEXT, DTC_CHOOSE]
F_ALL_IMAGES = "All images"
F_FILTERED_IMAGES = "Images matching a rule"
COL_PATH = "Path / URL"
COL_SERIES = "Series"
COL_INDEX = "Frame"
COL_URL = "FileLocation"
DEFAULT_METADATA_TAGS = (COL_URL, COL_SERIES, COL_INDEX)
IDX_EXTRACTION_METHOD_COUNT = 3
IDX_EXTRACTION_METHOD_COUNT_V1 = 1
IDX_EXTRACTION_METHOD_COUNT_V2 = 1
IDX_EXTRACTION_METHOD_COUNT_V3 = 1
IDX_EXTRACTION_METHOD = 4
IDX_EXTRACTION_METHOD_V1 = 2
IDX_EXTRACTION_METHOD_V2 = 2
IDX_EXTRACTION_METHOD_V3 = 2
LEN_EXTRACTION_METHOD_V1 = 8
LEN_EXTRACTION_METHOD = 9
CSV_JOIN_NAME = "CSV Metadata"
IPD_JOIN_NAME = "Image Metadata"
